#include <stdio.h>
#include <string.h>

#include "lrcmemory.h"
#include "ag_table.h"
#include "aoccpats.h"

void print_aocc_pattern(aocc_pattern_tp pat) {
  printf("Pattern:\n");
  printf("  kind: ");
  switch(pat.kind) {
    case AO_NORMAL : printf("normal"); break;
    case AO_LOCAL : printf("local"); break;
    case AO_NTA_SR : printf("nta"); break;
    case AO_ANY : printf("*"); break;
    default : printf("???");
  }
  printf("\n");
  printf("  production: %s\n", pat.prod_id);
  printf("  nonterminal: %s\n", pat.nt_id);
  printf("  multiplicity: %d\n", pat.multiplicity);
  printf("  attribute: %s\n", pat.attr_id);
}

#define streq(s0,s1) (!strcmp(s0,s1))
static aocc_pattern_tp construct_aocc_pattern(
   char * kind_id,
   char * prod_id, char * nt_id, int multiplicity, char * attr_id) {
  aocc_pattern_tp res;
  res.prod_id=prod_id;
  res.nt_id=nt_id;
  res.multiplicity=multiplicity;
  res.attr_id=attr_id;
  res.kind=AO_ANY;
  if (streq(kind_id,"local")) res.kind=AO_LOCAL;
  if (streq(kind_id,"normal")) res.kind=AO_NORMAL;
  if (streq(kind_id,"nta")) res.kind=AO_NTA_SR;
  return res;
}

/* substring returns a copy of the part of s that starts at s and
   ends before after.
 */
static char * substring(char * s, char * after) {
  char * res;
  int len;
  res=0;
  if (s==after) return s;
  len=strlen(s)-strlen(after);
  if (len<=0) return 0;
  /** printf("substring =%s=; =%s=; len=%d\n",s,after,len); **/
  res=(char *)ALLOC_STR(len+1);
  strncpy(res,s,len);
  res[len]=0;
  return res;
}

/* local attribute: P.a or *.a or P.* or *.* (a '.' and no ':')
   normal attribute: P:X$2.a, P:X.a or *-variants (a '.' and a ':' and possibly a '$')
   nta : P:X, P:X$2 or a *-variant ( a ':' and no '.')
 */
aocc_pattern_tp string_to_aocc_pattern(char * s) {
  char * dot_pos;
  char * dollar_pos;
  char * double_dot_pos;
  char * zero_pos;
  aocc_pattern_tp res;
  res=construct_aocc_pattern("*","*","*",-1,"*");

  double_dot_pos=strchr(s,':');
  dot_pos=strrchr(s,'.');
  dollar_pos=strrchr(s,'$');
  zero_pos=strrchr(s,0);
  if (!double_dot_pos && !dot_pos) {
    printf("Not a valid pattern: %s\n", s);
    return res;
  }
  if (!double_dot_pos && dot_pos) {
    printf("A local pattern: %s\n", s);
    res.kind=AO_LOCAL;
    res.prod_id=substring(s,dot_pos);
    res.attr_id=substring(dot_pos+1,zero_pos);
  }
  if (double_dot_pos && dot_pos) {
    printf("A normal pattern: %s\n", s);
    res.kind=AO_NORMAL;
    res.prod_id=substring(s,double_dot_pos);
    res.attr_id=substring(dot_pos+1,zero_pos);
    if (dollar_pos) {
      char * multipl_str; int multipl;
      res.nt_id=substring(double_dot_pos+1,dollar_pos);
      multipl_str=substring(dollar_pos+1,dot_pos);
      if (multipl_str && sscanf(multipl_str,"%d",&multipl)) res.multiplicity=multipl;
    }
    else { 
     res.nt_id=substring(double_dot_pos+1,dot_pos);
    }
  }
  if (double_dot_pos && !dot_pos) {  /* PROD:X, PROD:X$n */
    printf("A nta pattern %s\n", s);
    res.kind=AO_NTA_SR;
    res.prod_id=substring(s,double_dot_pos);
    if (dollar_pos) {
      char * multipl_str; int multipl;
      res.nt_id=substring(double_dot_pos+1,dollar_pos);
      multipl_str=substring(dollar_pos+1,zero_pos);
      if (multipl_str && sscanf(multipl_str,"%d",&multipl)) res.multiplicity=multipl;
    }
    else { res.nt_id=substring(double_dot_pos+1,zero_pos); }
  }
  /* some checks */
  if (!res.prod_id || !res.nt_id ||  !res.attr_id) {
    printf("Wrong format for pattern!\n");
  }
  if (!res.prod_id) res.prod_id="*";
  if (!res.nt_id) res.nt_id="*";
  if (!res.attr_id) res.attr_id="*";
  return res;
}

int aocc_matches_pattern(int aocc, aocc_pattern_tp pat) {
  int res;
  aocc_kind_tp ao_kind;
  int pat_has_prod, pat_has_nont, pat_has_attr, pat_has_multiplicity;

  pat_has_prod=!streq(pat.prod_id,"*");
  pat_has_nont=!streq(pat.nt_id,"*");
  pat_has_multiplicity=pat.multiplicity!=-1;
  pat_has_attr=!streq(pat.attr_id,"*");
  ao_kind=aocc_kind(aocc);

  if ((pat.kind==AO_ANY || pat.kind==AO_NORMAL) && aocc_kind(aocc)==AO_NORMAL) {
    res=!pat_has_prod || streq(pat.prod_id,prod_id(aocc_prod(aocc)));
    if (!res) return 0;

    res=!pat_has_nont || streq(pat.nt_id,ntocc_id(aocc_nocc(aocc)));
    if (!res) return 0;

    res=  !pat_has_multiplicity || pat.multiplicity==ntocc_multiplicity(aocc_nocc(aocc));
    if (!res) return 0;

    res=streq(pat.attr_id,"*") || streq(pat.attr_id,attr_id(aocc_attr(aocc))); 
    return res;
  }

  if (!pat_has_nont && !pat_has_multiplicity && 
      ((pat.kind==AO_ANY || pat.kind==AO_LOCAL) && aocc_kind(aocc)==AO_LOCAL)) {
    res=!pat_has_prod || streq(pat.prod_id,prod_id(aocc_prod(aocc)));
    if (!res) return 0;

    res=!pat_has_attr || streq(pat.attr_id,attr_id(aocc_attr(aocc))); 
    return res;
  }

  if (!pat_has_attr &&
      ( (pat.kind==AO_ANY || pat.kind==AO_NTA_SR) && aocc_kind(aocc)==AO_NTA_SR) ) {
    res=!pat_has_prod || streq(pat.prod_id,prod_id(aocc_prod(aocc)));
    if (!res) return 0;

    res=!pat_has_nont || streq(pat.nt_id,ntocc_id(aocc_nocc(aocc)));
    if (!res) return 0;

    res=  !pat_has_multiplicity || pat.multiplicity==ntocc_multiplicity(aocc_nocc(aocc));
    return res;
  }
  return 0;
}

void all_matching_equations(aocc_pattern_tp pat, void action(int)) {
  int eq_num, num_eqs;
  num_eqs=ag_num_equations();
  for(eq_num=0;eq_num<num_eqs;eq_num++) {
     int dest;
     dest=eq_destination(eq_num);
     if (aocc_matches_pattern(dest,pat)) {
       action(eq_num);
     }
  }
}

static void print_one_matching_equation(int eq_num) {
  printf("Equation %d matches ", eq_num); ag_print_aocc(eq_destination(eq_num));
  printf("\n");
}

void print_matching_equations(aocc_pattern_tp pat) {
  printf("Equations matching pattern\n");
  print_aocc_pattern(pat);
  all_matching_equations(pat,print_one_matching_equation);
}
