#include <stdlib.h>
#include <stdio.h>
#include "lrcassert.h"
#include "lrcmemory.h"
#include "strs.h"
#include "initial.h"
#include "grammar_tables.h"

typedef char * lrc_str_table[];
typedef int lrc_int_table[];

typedef struct transform_descriptor * p_transform_descriptor;

typedef struct transform_descriptor {
   p_transform_descriptor next;
   char * name;
   lrc_match_function_tp *matches_function;
   lrc_trafo_function_tp *transform_function;
} transform_descriptor;

typedef struct nont_info {
  char * ident;
  int num_prods;
  int entry_tok;
  char * scan_state;
  char * entry_attr;
  int first_visit_ix;
  int num_visits;
  int first_prod;
  lrc_term completing_term;
  lrc_term placeholder_term;
  p_transform_descriptor first_transform;
  p_transform_descriptor last_transform;
} nont_info_rec_tp;

typedef nont_info_rec_tp nont_info_table_tp[];
static nont_info_table_tp * nont_info_table;
#define NONT_FIELD(nt,f) (*nont_info_table)[nt].f

#define HEAD_TRAFOS(nt) NONT_FIELD(nt,first_transform)
#define TAIL_TRAFOS(nt) NONT_FIELD(nt,last_transform)

static int num_nonterminals;

typedef struct prod_info {
  char * ident;
  int num_sons;
  int lhs_nt;
  int same_lhs_prod;
  int first_visit_ix;
} prod_info_rec_tp;

typedef prod_info_rec_tp prod_info_table_tp[];
static prod_info_table_tp * prod_info_table;
#define PROD_FIELD(pr,f) (*prod_info_table)[pr].f

static int num_productions;
static int num_constructors;
static int root_nonterminal;

DebugDeclaration(static int num_allocs=0)

#ifdef DEBUG
#define is_prod(p) ((0<=(p)) && ((p)<num_constructors))
#define is_nont(n) ((0<=(n)) && ((n)<num_nonterminals))
#endif

static void nt_print_transformations(int nt) {
  p_transform_descriptor trafos;
  printf("Transformations for nonterminal %s\n", lrc_nont_ident(nt));
  trafos=HEAD_TRAFOS(nt);
  while (trafos!=0) {
    printf("  %s\n", trafos->name);
    trafos=trafos->next; 
  }
}

void lrc_alloc_grammar_tables(int num_nts, int num_prods, int num_wrappers) {

 Assert(num_allocs==0,"wrong number of grammar table allocations\n");
 DebugStatement(num_allocs++)
 num_nonterminals=num_nts;
 nont_info_table=(nont_info_table_tp *)lrc_malloc(num_nts * sizeof((*nont_info_table)[0]));
 
 num_productions=num_prods;
 num_constructors=num_productions+num_wrappers;
 
 prod_info_table
 =(prod_info_table_tp *)lrc_malloc(num_constructors * sizeof((*prod_info_table)[0]));
   { int ix;
     for (ix=0;ix<num_constructors; ix++) {
       PROD_FIELD(ix,ident)=0;
       PROD_FIELD(ix,num_sons)=-1;
       PROD_FIELD(ix,lhs_nt)=-1;
       PROD_FIELD(ix,same_lhs_prod)=-1;
       PROD_FIELD(ix,first_visit_ix)=-1;
     }
   }
}

void lrc_print_prod_info(int prod) {
  printf("Production %d\n", prod);
  printf("    ident=%s\n",PROD_FIELD(prod,ident));
  printf("    num_sons=%d\n",PROD_FIELD(prod,num_sons));
  printf("    lhs_nt=%d\n",PROD_FIELD(prod,lhs_nt));
  printf("    same_lhs_prod=%d\n",PROD_FIELD(prod,same_lhs_prod));
  printf("    first_visit_ix=%d\n",PROD_FIELD(prod,first_visit_ix));
}

/* This routine defines that vpr_num is the number of a visit-routine
 * for production prod. We only record the first of these, since these
   routines are numbered consecutively.
 */
void lrc_define_prod_visit_index(int prod, int vpr_num) {
  AssertI(is_prod(prod),"lrc_define_prod_visit_index:prod range error",prod)
  if (PROD_FIELD(prod,first_visit_ix)==-1) PROD_FIELD(prod,first_visit_ix)=vpr_num;
}

static void nont_define_first_prod(int nont, int prod) {
  if (nont==-1) return; /* wrappers have no lhs */
  AssertI(is_nont(nont),"nont_define_first_prod:nont range error",nont)
  AssertI(is_prod(prod),"nont_define_first_prod:prod range error",prod)
  NONT_FIELD(nont,first_prod)=prod;
}

void lrc_define_nonterminal(
     char * ident, int global_number, int num_prods,
     int entry_tok, char * scan_state, char * entry_attr,
     int first_visit_ix, int num_visits
  ) {
  Assert(num_allocs==1,"lrc_define_nonterminal")
  AssertI(is_nont(global_number),"lrc_define_nonterminal:range error",global_number)
  NONT_FIELD(global_number,ident)=ident;
  NONT_FIELD(global_number,num_prods)=num_prods;
  NONT_FIELD(global_number,entry_tok)=entry_tok;
  NONT_FIELD(global_number,scan_state)=scan_state;
  NONT_FIELD(global_number,entry_attr)=entry_attr;
  NONT_FIELD(global_number,first_visit_ix)=first_visit_ix;
  NONT_FIELD(global_number,num_visits)=num_visits;
  NONT_FIELD(global_number,first_prod)=-1;
  NONT_FIELD(global_number,first_transform)=0;
  NONT_FIELD(global_number,last_transform)=0;
}

void lrc_nont_set_completing_term(int nt, lrc_term trm) {
  NONT_FIELD(nt,completing_term)=trm;
}

lrc_term lrc_nont_get_completing_term(int nt) {
  return NONT_FIELD(nt,completing_term);
}

void lrc_nont_set_placeholder_term(int nt, lrc_term trm) {
  NONT_FIELD(nt,placeholder_term)=trm;
}

lrc_term lrc_nont_get_placeholder_term(int nt) {
  return NONT_FIELD(nt,placeholder_term);
}

void lrc_define_root(int r) {
  AssertI(is_nont(r),"lrc_define_root:range error",r)
  root_nonterminal=r;
}

void lrc_print_nont_info(int nt) {
  printf("Nonterminal %d\n",nt);
  printf("    ident=%s\n",NONT_FIELD(nt,ident));
  printf("    num_prods=%d\n",NONT_FIELD(nt,num_prods));
  printf("    entry_tok=%d\n",NONT_FIELD(nt,entry_tok));
  printf("    scan_state=%s\n",NONT_FIELD(nt,scan_state));
  printf("    entry_attr=%s\n",NONT_FIELD(nt,entry_attr));
  printf("    first_visit_ix=%d\n",NONT_FIELD(nt,first_visit_ix));
  printf("    num_visits=%d\n",NONT_FIELD(nt,num_visits));
  printf("    first_prod=%d\n",NONT_FIELD(nt,first_prod));
}

/* Access to NONTERMINAL information */

int lrc_root_nonterminal() {
  return root_nonterminal;
}

int lrc_nont_num_prods(int nt) {
  AssertI(is_nont(nt),"lrc_nont_num_prods:range error",nt)
  return NONT_FIELD(nt,num_prods);
}

int lrc_nont_entry_token(int nt) {
  AssertI(is_nont(nt),"lrc_nont_entry_token:range error",nt)
  return NONT_FIELD(nt,entry_tok);
}

char * lrc_nont_ident(int nt) {
  AssertI(is_nont(nt),"lrc_nont_ident:range error",nt)
  return NONT_FIELD(nt,ident);
}

char * lrc_nont_scan_state(int nt) {
  AssertI(is_nont(nt),"lrc_nont_scan_state:range error",nt)
  return NONT_FIELD(nt,scan_state);
}

int lrc_nont_first_visit_index(int nt) {
  AssertI(is_nont(nt),"lrc_nont_first_visit_index:range error",nt)
  return NONT_FIELD(nt,first_visit_ix);
}

int lrc_nont_num_visits(int nt) {
  AssertI(is_nont(nt),"lrc_nont_num_visits:range error",nt)
  return NONT_FIELD(nt,num_visits);
}

int lrc_nont_first_prod(int nt) {
  AssertI(is_nont(nt),"lrc_nont_first_prod:range error",nt)
  return NONT_FIELD(nt,first_prod);
}

int lrc_id_to_nont(char * nt_id) {
  int index;
  
  for (index=0; index<num_nonterminals; ++index) {
    if (lrc_string_compare(nt_id,lrc_nont_ident(index))==lrc_equal) {
      return index;
    }
  }
  /*
  ** None was found. Return an error value.
  */
  return -1;
}

/* Defining constructors */

void lrc_define_constructor(
  char * ident, int global_number, int num_sons, int lhs_nt) {
  Assert(num_allocs==1,"lrc_define_constructor")
  AssertI(is_prod(global_number),"lrc_define_constructor:range error",global_number)
  PROD_FIELD(global_number,ident)=ident;
  PROD_FIELD(global_number,num_sons)=num_sons;
  PROD_FIELD(global_number,lhs_nt)=lhs_nt;
  if (lhs_nt!=-1) PROD_FIELD(global_number,same_lhs_prod)=NONT_FIELD(lhs_nt,first_prod);
  nont_define_first_prod(lhs_nt,global_number);
}

/* Access to constructor information */
int lrc_num_constructors() {
  Assert(num_allocs==1,"lrc_num_constructors")
  return num_constructors;
}

int lrc_num_productions() {
  Assert(num_allocs==1,"lrc_num_productions")
  return num_productions;
}

char * lrc_constructor_ident(int constructor_number) {
  Assert(num_allocs==1,"lrc_constructor_ident")
  AssertI(is_prod(constructor_number),"lrc_constructor_ident:range error",constructor_number)
  Assert(PROD_FIELD(constructor_number,ident)!=0,"lrc_constructor_ident:undefined entry")
  return PROD_FIELD(constructor_number,ident);
}

int lrc_constructor_num_sons(int constructor_number) {
  Assert(num_allocs==1,"lrc_constructor_num_sons")
  /** Assert(is_prod(constructor_number),"lrc_constructor_num_sons:range error") **/
  if (!is_prod(constructor_number)) {
     printf("?????");
     Assert(is_prod(constructor_number),"lrc_constructor_num_sons:range error")

  }
  Assert(PROD_FIELD(constructor_number,num_sons)!=-1,"lrc_constructor_num_sons:undefined entry")
  return PROD_FIELD(constructor_number,num_sons);
}

int lrc_prod_lhs_nont(int constructor_number) {
  AssertI(is_prod(constructor_number),"lrc_prod_lhs_nont:range error",constructor_number)
  return PROD_FIELD(constructor_number,lhs_nt);
}

int lrc_prod_next_prod(int constructor_number) {
  AssertI(is_prod(constructor_number),"lrc_prod_next_prod:range error",constructor_number)
  return PROD_FIELD(constructor_number,same_lhs_prod);
}

int lrc_prod_first_visit_index(int prod) {
  AssertI(is_prod(prod),"lrc_prod_first_visit_index:range error",prod)
  return PROD_FIELD(prod,first_visit_ix);
}

int lrc_prod_num_visits(int prod) {
  return lrc_nont_num_visits(lrc_prod_lhs_nont(prod));
}

/* 
** Added by Dick Heijink
*/

/*
** When a module is part of the runtime library, it can not use constant
** values for the tags. These values might change from grammar to grammar.
** The next function can retrieve the value of the tag based on the name
** of the production. The function returns a tagvalue of -1 when the
** name was not found among the productions of the grammar.
** Precondition: True
*/
int lrc_prod_name_to_tag(lrc_string name)
{
  int index;
  
  /*
  ** Scan all production, one by one, and determine whether the name of the 
  ** production equals the name of the argument.
  */
  for ( index=0; index<num_constructors; ++index ) {
    if ( lrc_string_compare(name,lrc_constructor_ident(index))==lrc_equal ) {
      return index;
    }
  }
  /*
  ** None was found. Return an error value.
  */
  return -1;
}

static p_transform_descriptor create_transform_descriptor(
  char * tr_name, lrc_match_function_tp m_fun, lrc_trafo_function_tp tr_fun) {
  p_transform_descriptor result;
  result=(p_transform_descriptor)lrc_malloc(sizeof(*result));
  result->name=tr_name;
  result->next=0;
  result->matches_function=m_fun;
  result->transform_function=tr_fun;
  return result;
}


void lrc_define_transformation(
  char * nt_id, char * tr_name, lrc_match_function_tp m_fun, lrc_trafo_function_tp tr_fun) {
   int nt;

   /* printf("Defining transformation %s for %s\n", tr_name, nt_id);
    */
   nt=lrc_id_to_nont(nt_id);
   if (nt==-1) printf(" no such nonterminal %s????\n", nt_id);
   if (nt!=-1) {
      p_transform_descriptor descriptor;
      descriptor=create_transform_descriptor(tr_name,m_fun,tr_fun);
      if (HEAD_TRAFOS(nt)==0) {
         HEAD_TRAFOS(nt)=descriptor;
         TAIL_TRAFOS(nt)=descriptor;
      }
      else {
        TAIL_TRAFOS(nt)->next=descriptor;
        TAIL_TRAFOS(nt)=descriptor;
      }
   /* nt_print_transformations(nt);
    */
   }
}

static p_transform_descriptor lookup_trafo_descriptor(
   char * tr_name, lrc_term t, p_transform_descriptor trafos) {
   if (trafos==0) return 0;
   if (lrc_string_compare(tr_name,trafos->name)==lrc_equal
       && (trafos->matches_function)(t)
      )
      return trafos;
   return lookup_trafo_descriptor(tr_name,t,trafos->next);
}

static lrc_trafo_function_tp * lrc_transformation_function(lrc_term t, char * trafo_name) {
  int nt;
  p_transform_descriptor td;

  nt=lrc_prod_lhs_nont(LRC_TAG(t));
  td=lookup_trafo_descriptor(trafo_name,t,HEAD_TRAFOS(nt));
  if (td==0) return 0;
  return td->transform_function;
}

lrc_term lrc_apply_transformation(lrc_term trm, char * trafo_name) {
  lrc_trafo_function_tp * trafo_function;
  lrc_term shadow_trm;
  lrc_term replacement;

  shadow_trm=lrc_shadow(trm);
  trafo_function=lrc_transformation_function(shadow_trm,trafo_name);
  if (trafo_function==0) {
    printf("lrc_apply_transformation: no function for '%s'\n", trafo_name);
    return shadow_trm;
  }
  replacement=(*trafo_function)(shadow_trm);
  /*
  printf("lrc_apply_transformation: replacement of\n  ");
  lrc_term_print_long(trm);
  printf("  is\n  ");
  lrc_term_print_long(replacement);
  */
  return replacement;
}

p_lrc_string_list lrc_strings_2(lrc_string elem, p_lrc_string_list lst) {
  p_lrc_string_list res;
  res=(p_lrc_string_list)lrc_malloc(sizeof(*res));
  res->next=lst;
  res->str=elem;
  return res;
}

static p_lrc_string_list remove_str(char * s, p_lrc_string_list strs) {
  if (strs==0) return 0;
  if (lrc_string_compare(s,strs->str)==lrc_equal) return strs->next;
  return lrc_strings_2(strs->str,remove_str(s,strs->next));
}

int lrc_string_in_list(char * s, p_lrc_string_list strs) {
  if (strs==0) return 0;
  if (lrc_string_compare(s,strs->str)==lrc_equal) return 1;
  return lrc_string_in_list(s,strs->next);
}


p_lrc_string_list lrc_matches_from_list(lrc_term t, p_transform_descriptor trafos) {
  lrc_bool match_result;
  p_lrc_string_list rest_matches;
  if (trafos==0) return 0;
  match_result=(trafos->matches_function)(t);

  /*
  printf(" trafo %s ", trafos->name);
  if (match_result) { printf("matches\n"); }
  else { printf("does not match\n"); }
  */

  rest_matches=lrc_matches_from_list(t,trafos->next);
  if (match_result) {
     return lrc_strings_2(trafos->name,remove_str(trafos->name,rest_matches));
  }
  return rest_matches;
}

p_lrc_string_list lrc_matching_transformations(lrc_term t) {
  int nt;
  nt=lrc_prod_lhs_nont(LRC_TAG(t));
  /* nt_print_transformations(nt);
   */
  return lrc_matches_from_list(lrc_shadow(t),HEAD_TRAFOS(nt));
}
