/*
** shalloc.c
** january 1994 by Maarten Pennings
** december 1994 (Matthijs Kuiper)
**   pointers point to first field
**   garbage collection
**     added mark field
** november 1995 (Matthijs Kuiper)
**   added switch FAT
**   merged size and mark fields
** january 1998 (Matthijs Kuiper)
**   splitted shalloc.c into shalloc.c and term.c
*/

#include <memory.h>
#include <stdio.h>
#include "data_layout.h"
#include "shalloc.h"
#include "lrcmemory.h"
#include "prims.h"

#define FAIL          0
#define ALLOC_CELL(n) lrc_malloc(n)

static lrc_ptr tabel[lrc_term_tabelsize];
static int stats_num_allocs;
static int stats_num_hits;
static int stats_size_allocs;

void lrc_term_init() {
  lrc_term_ix i;
  for(i=0; i<lrc_term_tabelsize; i++) tabel[i]=FAIL;
  stats_num_allocs=0;
  stats_num_hits=0;
  stats_size_allocs=0;
  if (sizeof(lrc_ptr)!=sizeof(int) || sizeof(lrc_term_ix)!=sizeof(int)) {
    printf("You're in trouble.\n");
    printf("This library expects certain types to have the same size.\n");
    exit(1);
  }
  if (sizeof(short_bits_tp)!=sizeof(int)) {
    printf("You're in trouble.\n");
    printf("This library expects certain types to have the same size.\n");
    printf("On this machine short_bits_tp and int have different sizes.\n");
    exit(1);
  }
  return;
}


static lrc_ptr lrc_term_register_miss(lrc_term_ix i,int size,lrc_ptr buf) {
  lrc_ptr *t=(lrc_ptr*)ALLOC_CELL(CELL_SIZE(size));
  lrc_ptr res=t;
  short n_nodes = 1;
  *t++=tabel[i];       /* NEXT */
#ifdef FAT
  { int n,tg; 
    int n_sons = size / sizeof(lrc_ptr) - 1;
    lrc_ptr *sons = (lrc_ptr *) buf;
    tg  = (int) *(sons);
/*fprintf(stdout,"tg %d ",(int) *(sons+0));fflush(stdout);*/
    if ( tg >= lrc_first_free_tag )    
      for ( n = 1 ; n <= n_sons ; n++ )
	{ n_nodes += NNOD(sons[n]);
/*	  fprintf(stdout," n_nodes: %d  ",n_nodes);fflush(stdout); */
	}
  } 
  MARK(res) = (short) 0;
  NNOD(res) = (short) n_nodes;
  t++;
  *t++=(lrc_ptr)i;     /* HASH */
  *t++=(lrc_ptr)0;     /* HITS */
  *t++=(lrc_ptr)stats_num_allocs; /* BIRTH */
  *t++=(lrc_ptr)size;  /* SIZE */
#else
  *t++=(lrc_ptr)i;     /* HASH */
  SIZE(res)=size;
  CLEAR_MARK(res);
  *t++;
#endif
  CLEAR_INITIAL(res);
  tabel[i]=res;
  memcpy(t,buf,size);
#ifdef DEBUG
  if ((*(int *)buf)<0) {
    printf("strange tag in lrc_term_register_miss: %d\n",*(int *)buf);
  }
#endif
  stats_num_allocs+=1;
  stats_size_allocs+=CELL_SIZE(size);
  return res;
}

void lrc_term_register_hit(lrc_ptr t) { 
  stats_num_hits++;
#ifdef FAT
  HITS(t)++;
  BIRTH(t)=stats_num_allocs; /* change the birth-time to reflect hit */
#endif
  return;
}

#define INSERT(t,i,n,buf) \
  if (t==FAIL) t=lrc_term_register_miss(i,n,buf); else lrc_term_register_hit(t);

#define REHASH(ix1,ix2) \
 ((2*(ix1)+(ix2)) % lrc_term_tabelsize)

lrc_ptr lrc_shalloc(lrc_term_ix i,lrc_selector eq,int size,int sons,lrc_ptr buf) {
  lrc_ptr t,*son=(lrc_ptr*)((int)buf+size-sons*sizeof(lrc_ptr));
  { int n; for(n=0; n<sons; n++) i=REHASH(i,HASH(son[n])); }
  for(t=tabel[i]; t!=FAIL && !eq(SIZE(t),t,size,sons,buf); t=NEXT(t));
  INSERT(t,i,size,buf); 
  return t;
}

/************************\
** High level interface **
\************************/

/*
**  FAT interface
**       0     1      2      3     4       5     6
**   +------+------+------+------+------+------+------+------ ... ------+
**   | next | mark | hash | hits | birth| size |  tag   primitive value |
**   +------+------+------+------+------+------+------+------ ... ------+
**
**   +------+------+------+------+------+------+------+------+---- ... ----+------+
**   | next | mark | hash | hits | birth| size |  tag      pointers to subterms   |
**   +------+------+------+------+------+------+------+------+---- ... ----+------+

**  LEAN interface
**      0      1       2          3
**   +------+------+----------+------+------ ... ------+
**   | next | hash |mark+size |  tag   primitive value |
**   +------+------+----------+------+------ ... ------+
**
**   +------+------+-----------+------+----------------- ... ------+
**   | next | hash | mark+size |  tag  pointers to subterms        |
**   +------+------+-----------+------+----------------- ... ------+
**
*/

lrc_term lrc_shared_prim(lrc_term_ix i,lrc_comparer eq,int size,lrc_ptr buf) {
  lrc_term t;
  for(t=tabel[i]; t!=FAIL && (SIZE(t)!=size || !eq(t,buf)); t=NEXT(t));
  INSERT(t,i,size,buf);
  SET_PRIMITIVE(t);
  return t;
}

#ifdef TRACE
static void (*term_tracer)(lrc_term);
static int term_tracer_defined=0;

void define_term_tracer(void tracer()) {
  term_tracer=tracer;
  term_tracer_defined=(tracer!=0);
}
#endif

#define TERM_INDEX_OFFSET 3000
#define HASH_LRC_TAG(tag) (tag+TERM_INDEX_OFFSET)%lrc_term_tabelsize

lrc_term lrc_shared_cons(int size,lrc_term buf) {
  lrc_term t; int num=size/sizeof(lrc_ptr);
  int i=HASH_LRC_TAG(LRC_TAG(buf-ADMIN_SIZE)); /* terrible hack: buf-ADMIN_SIZE !! */
  { int n; for(n=1; n<num; n++) i=REHASH(i,HASH(buf[n])); }
  for(t=tabel[i]; t!=FAIL && (SIZE(t)!=size || memcmp(t+ADMIN_SIZE,buf,size)!=0); t=NEXT(t));
  INSERT(t,i,size,buf);
#ifdef TRACE
  if (term_tracer_defined) term_tracer(t);
#endif
  return t;
}

/***************\
** Diagnostics **
\***************/

int  lrc_term_time() {
  return stats_num_allocs;
}

int  lrc_term_hashix(lrc_term t) {
  return HASH(t);
}

int lrc_term_nnodes(lrc_term t)
{
#ifdef FAT
  return NNOD(t);
#else
  return 0;
#endif
}


int  lrc_term_hits(lrc_term t) {
#ifdef FAT
  return HITS(t);
#else
  return 0;
#endif
}

int  lrc_term_birth(lrc_term t) {
#ifdef FAT
  return BIRTH(t);
#else
  return 0;
#endif
}

int lrc_term_size(lrc_term t) {
  return SIZE(t);
}

int lrc_term_occupied_size(lrc_term t) {
  return SIZE(t) + ADMIN_SIZE*sizeof(lrc_ptr);
}

int lrc_term_mark(lrc_term t) {
  return MARK(t);
}

static int birth_base=0;
void lrc_term_dump_set_birth_time(int birth_time) {
  birth_base=birth_time;
  return;
}

void lrc_all_terms_in_cache(void action(lrc_term)){
  lrc_term_ix i; 
  for(i=0; i<lrc_term_tabelsize; i++) {
    lrc_ptr t;
    for(t=tabel[i]; t!=FAIL; t=NEXT(t)) action((lrc_term)t);
  }
  return;
}

void lrc_term_dump(lrc_term_filter filter,lrc_term_printer printer) {
  lrc_term_ix i; 
  printf("DUMP TERM base= %d\n\n",birth_base);
  for(i=0; i<lrc_term_tabelsize; i++) {
    lrc_ptr t; int n=0; int m=0;
    for(t=tabel[i]; t!=FAIL; t=NEXT(t),n++) 
      if (filter((lrc_term)t) && lrc_term_birth((lrc_term)t)>=birth_base) {
        printf("  TERM %d %d ",lrc_term_hits(t),lrc_term_birth(t));
        printer((lrc_term)t); 
        printf("\n");
      } else 
        m++;
    if (n>m) printf("BUCKET %d  list= %d  hide= %d  TOTAL= %d\n\n",i,n-m,m,n);
  }
  return;
}

static float root(float x) {
  float a=0,b=x,m;
  while ((a-b)*(a-b)>1E-08) { m=(a+b)/2; if (m*m<x) a=m; else b=m; }
  return (a+b)/2;
}

static void stats(int lev,float *aver,float *sdev,int *longest,int *shortest) {
  lrc_term_ix i; float av,var; int min; int max;

  av=(float)stats_num_allocs/(float)lrc_term_tabelsize; 
  var=0; min=stats_num_allocs; max=0;

  for(i=0; i<lrc_term_tabelsize; i++) {
    lrc_ptr t; int n;
    for(t=tabel[i],n=0; t!=FAIL; t=NEXT(t),n++) ;
    if (lev>3) printf("%d : %d\n",i,n);
    var+=(av-n)*(av-n);
    if (n>max) max=n;
    if (n<min) min=n;
  }

  var=var/(float)(lrc_term_tabelsize-1); 
  *aver=av; *sdev=root(var); *longest=max; *shortest=min;

  return;
}

void lrc_term_stats(int level) {
  /* level 0: hits/misses
  ** level 1: admin/user size
  ** level 2: min/max bucketlength
  ** level 3: normal distribution
  ** level 4: each bucket: length
  */
  float av; float sd; int longest; int shortest;

  printf("TERM   n=%d  misses=%d  hits=%d  total=%d\n",
    lrc_term_tabelsize, stats_num_allocs,
    stats_num_hits,stats_num_allocs+stats_num_hits);
  if (level==0) return;

  printf("SIZE   total=%d  admin=%d  user=%d\n",
    stats_size_allocs,
    stats_num_allocs*ADMIN_SIZE*sizeof(lrc_ptr),
    stats_size_allocs-stats_num_allocs*ADMIN_SIZE*sizeof(lrc_ptr));
  if (level==1) return;

  stats(level,&av,&sd,&longest,&shortest);
  printf("LENGTH Min=%d  Av=%.1f  Max=%d  Sd=%.2f\n",shortest,av,longest,sd);
  if (level==2) return;

  printf("84%% of the lists is shorter than %.1f (=Av+%.1f%%)\n"
        ,av+sd,sd/av*100);
  printf("98%% of the lists is shorter than %.1f (=Av+%.1f%%)\n"
        ,av+2*sd,2*sd/av*100);
  /* in a normal distribution, 68% lies in av+/-sd  */
  /*                       and 95% lies in av+/-2sd */
  return;
}

/* marking and deleting entries from the constructor cache
 */

static lrc_ptr remove_unmarked_in_list(lrc_ptr l) {
  if (l==FAIL) return l;
  if (MARK(l)) {
    NEXT(l)=remove_unmarked_in_list(NEXT(l));
    CLEAR_MARK(l);
    return l;
  }
  return remove_unmarked_in_list(NEXT(l));
}

void remove_unmarked_terms() {
  lrc_term_ix i;
  for(i=0; i<lrc_term_tabelsize; i++) {
    tabel[i]=remove_unmarked_in_list(tabel[i]);
  }
}
