/***	textbuffer.c   -   Matthijs Kuiper - june 1996

	implementation of textbuffers
***/

#include <stdio.h>
#include <string.h>
#include "lrcmemory.h"
#include "textbuffer.h"

#define ALLOC_UNIT	128


Textbuffer textbuffer_init(void)
{
	Textbuffer tb;

	tb=(Textbuffer)lrc_malloc(sizeof(*tb));
	if (tb==NULL)
	{	fprintf(stderr,"malloc error in textbuffer_init\n");
		return NULL;
	}
	tb->content=NULL;
	tb->size=0;
	tb->free=0;
	return tb;
}

void textbuffer_clear_content(Textbuffer tb) {
  tb->free=tb->size;
  if (tb->content!=0) {
    tb->content[0]='\0';
    /* int ix;
      for(ix=0;ix<tb->size;ix++) tb->content[ix]='\0';
    */
  }
}

static void textbuffer_grow_space(Textbuffer tb) {
  if (tb->content==0) {
    tb->content=(char *)ALLOC_STR(ALLOC_UNIT);
    tb->size=ALLOC_UNIT;
    tb->free=tb->size;
    return;
  }
  /*** printf("text_buffer_grow_space --> %d\n", 2*(tb->size) ); ***/
  tb->content=(char *)lrc_realloc(tb->content,(size_t)(2*(tb->size)));
  tb->free=tb->free+tb->size;
  tb->size=2*(tb->size);
}

/* report the length of the largest string we can accomodate,
   INCLUDING its final zero-byte
 */
static int tb_available_string_space(Textbuffer tb) {
  if (tb->content==0) return 0;
  if (tb->size==0) return 0;
  if (tb->free==tb->size) return tb->free;
  return tb->free+1;
}

/* report the number of bytes that are used in contents */
int textbuffer_used_size(Textbuffer tb) {
   return tb->size - tb_available_string_space(tb);
}

/* assure that the free space in tb can hold the additional number of bytes s */
static void textbuffer_assure_free_space(Textbuffer tb, int s) {
  while(tb_available_string_space(tb)<s) textbuffer_grow_space(tb);
}

static int tb_index_in_range(Textbuffer tb, int index) {
  int min_index, max_index; /* allowable indices are in [min_index,max_index) */
  if (tb->content==0) { min_index=0; max_index=0; }
  else if (tb->size==tb->free) { min_index=0; max_index=tb->size; }
  else { min_index=tb->size-tb->free-1;  max_index=tb->size;}
  return (min_index<=index) && (index<max_index);
}

/***** only used in PARANOID version of text_buffer_append_string 
static void tb_set_char(Textbuffer tb, int index, char c) {
  if (!tb_index_in_range(tb,index)) {
     printf("Index out of range in tb_set_char index=%d;size=%d;free=%d\n",
             index, tb->size, tb->free);
     fflush(stdout);
     return;
  }
  tb->content[index]=c;
}
**********/

/* this function adds the string s to the textbuffer tb.
 * Precondition: the textbuffer has enough space to hold string s
 *
 */
static void textbuffer_append_string(Textbuffer tb, char * s) {
  if (tb->size==tb->free) {
    tb->content[0]='\0';
    tb->content=strcat(tb->content,s);
    tb->free=tb->free-(strlen(s)+1);
    return;
  } else {
    int len, start_index;
    len=strlen(s);
    if (tb->size==tb->free) { start_index=0; }
    else { start_index=(tb->size-tb->free)-1; }
    strcpy((tb->content)+start_index,s);
/*** PARANOID Version of copying s into t->content:
    {
      int index, last_index;
      last_index=start_index+len;
      if (tb->content[start_index]!='\0') {
        printf("textbuffer_append_string:Content[start_index=%d] is not a zero-byte???\n",
                start_index);
        fflush(stdout);
      }
      for (index=0;index<=len;index++) {
        tb_set_char(tb,start_index+index,s[index]);
      }
      if (tb->content[last_index]!='\0') {
        printf("textbuffer_append_string:Content is not terminated by a zero-byte???\n");
        fflush(stdout);
      }
    }
***/
    tb->free=tb->free-len;
  }
}

void textbuffer_append_char(Textbuffer tb, char c) {
  textbuffer_assure_free_space(tb,2);
  if (tb->size==tb->free) {
    tb->content[0]=c;
    tb->content[1]='\0';
    tb->free=tb->free-2;
  } else {
    int start_index;
    start_index=(tb->size-tb->free)-1;
    tb->content[start_index]=c;
    tb->content[start_index+1]='\0';
    tb->free=tb->free-1;
  }
}

void textbuffer_append_n_chars(Textbuffer tb, int n, char c) {
  if (n==0) return;
  textbuffer_assure_free_space(tb,n+1);
  if (tb->size==tb->free) {
    memset(tb->content,c,n);
    tb->content[n]='\0';
    tb->free=tb->free-(n+1);
    return;
  } else {
    int start_index;
    start_index=(tb->size-tb->free)-1;
    memset(tb->content+start_index,c,n);
    tb->content[start_index+n]='\0';
    tb->free=tb->free-n;
  }
}

void textbuffer_append(Textbuffer tb,char *s) {
  textbuffer_assure_free_space(tb,strlen(s)+1);
  textbuffer_append_string(tb,s);
}

void textbuffer_store(Textbuffer tb,char *s)
{  textbuffer_clear_content(tb);
   textbuffer_append(tb,s);
}

void textbuffer_destroy(Textbuffer tb) {
  if (tb==0) return;
  lrc_free(tb->content);
  lrc_free(tb);
}
