#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include "lrcassert.h"
#include "lrcmemory.h"
#include "strs.h"
#include "basics.h"
#include "prims.h"
#include "memo.h"
#include "grammar_tables.h"
#include "visit_tables.h"

typedef struct packet_attrs {
  struct packet_attrs * next;
  char *                name;
  int                   index;
} *PacketAttrs;

typedef struct visit_info {
  lrc_ptr     routine;
  int         global_number;
  int         production_number;
  int         visit_number;
  int         num_ip_attrs;
  int         num_op_attrs;
  PacketAttrs ip_attrs;
  PacketAttrs op_attrs;
}  visit_info;

typedef visit_info lrc_visit_table[];
static lrc_visit_table * visit_table;

typedef struct nont_visit_info {
  lrc_ptr     routine;
  int         nont_number;
  int         visit_number;
  int         num_ip_attrs;
  int         num_op_attrs;
}  nont_visit_info;

typedef nont_visit_info lrc_nont_visit_table[];
static lrc_nont_visit_table * nont_visit_table;

static int num_vnt_routines;
static int num_vpr_routines;

#ifdef DEBUG
static int is_vnt_index(int vnt_ix) {
  return (0<=vnt_ix) && (vnt_ix<num_vnt_routines);
}

static int is_vpr_index(int vpr_ix) {
  return (0<=vpr_ix) && (vpr_ix<num_vpr_routines);
}
#endif

DebugDeclaration(static int num_allocs=0)

void lrc_alloc_visit_table(int num_vprs, int num_vnts) {
 Assert(num_allocs==0,"wrong number of visit table allocations")
 DebugStatement(num_allocs++)

 visit_table=(lrc_visit_table *)lrc_malloc(num_vprs*sizeof(visit_info));
 nont_visit_table=(lrc_nont_visit_table *)lrc_malloc(num_vnts*sizeof(nont_visit_info));

 Assert(visit_table!=0,"allocation of production visit table failed")
 Assert(nont_visit_table!=0,"allocation ofnonterminal visit table failed")
 num_vpr_routines=num_vprs;
 num_vnt_routines=num_vnts;
}

void lrc_define_visit(void vpr_routine(), int global_number, int prod_number, int vis_number,
      int num_ip_attrs, int num_op_attrs) {
  Assert(num_allocs==1,"visit table not allocated")
  AssertI(is_vpr_index(global_number),
          "lrc_define_visit:range error:global_number",global_number)
  AssertI(0<vis_number,"lrc_define_visit:non-positive visit number",vis_number)
  AssertI(vis_number<=lrc_prod_num_visits(prod_number),
          "lrc_define_visit:range error:vis_number",vis_number)
  (*visit_table)[global_number].routine           = (lrc_ptr)vpr_routine;
  (*visit_table)[global_number].production_number = prod_number;
  (*visit_table)[global_number].visit_number      = vis_number;
  (*visit_table)[global_number].num_ip_attrs      = num_ip_attrs;
  (*visit_table)[global_number].num_op_attrs      = num_op_attrs;
  (*visit_table)[global_number].ip_attrs          = 0;
  (*visit_table)[global_number].op_attrs          = 0;
  lrc_define_prod_visit_index(prod_number, global_number);
}

void lrc_define_nont_visit(
  void vnt_routine(), int global_number, int nont_number,
  int visit_number, int num_ip_attrs, int num_op_attrs) {
  AssertI(is_vnt_index(global_number),
          "lrc_define_nont_visit:range error:global_number",global_number)
  AssertI(0<visit_number,"lrc_define_nont_visit:non-positive visit_number",visit_number)
  AssertI(visit_number<=lrc_nont_num_visits(nont_number),
          "lrc_define_nont_visit:range error:visit_number",visit_number)
  (*nont_visit_table)[global_number].routine      = (lrc_ptr)vnt_routine;
  (*nont_visit_table)[global_number].nont_number  = nont_number;
  (*nont_visit_table)[global_number].visit_number = visit_number;
  (*nont_visit_table)[global_number].num_ip_attrs = num_ip_attrs;
  (*nont_visit_table)[global_number].num_op_attrs = num_op_attrs;
}


static PacketAttrs enter_attr_description(PacketAttrs pas, char * attr_name, int packet_index) {
  PacketAttrs new_elem;
  new_elem=(PacketAttrs)lrc_malloc(sizeof(*new_elem));
  new_elem->name=attr_name;
  new_elem->index=packet_index;
  new_elem->next=pas;

  return new_elem;
}

static int lookup_in_packet_attrs(PacketAttrs pas, char * attr_name) {
  while (pas!=0) {
     if (lrc_string_equal(attr_name,pas->name)) {
        return pas->index;
     }
     pas=pas->next;
  }
  return -1;
}

static char * lookup_index_in_packet_attrs(PacketAttrs pas, int packet_index) {
  while (pas!=0) {
     if (packet_index==pas->index) {
        return pas->name;
     }
     pas=pas->next;
  }
  return 0;
}

char * lrc_attr_get_ip_name(int nt_num, int vis_num, int packet_index ) {
  int global_vnum;
  global_vnum=lrc_prod_visnum_vpr_num(lrc_nont_first_prod(nt_num), vis_num);
  return lookup_index_in_packet_attrs((*visit_table)[global_vnum].ip_attrs,packet_index);
}

char * lrc_attr_get_op_name(int nt_num, int vis_num, int packet_index ) {
  int global_vnum;
  global_vnum=lrc_prod_visnum_vpr_num(lrc_nont_first_prod(nt_num), vis_num);
  return lookup_index_in_packet_attrs((*visit_table)[global_vnum].op_attrs,packet_index);
}

int lrc_attr_get_ip_index(int nt_num, int vis_num, char * attr_name) {
  int global_vnum;
  global_vnum=lrc_prod_visnum_vpr_num(lrc_nont_first_prod(nt_num), vis_num);
  return lookup_in_packet_attrs((*visit_table)[global_vnum].ip_attrs,attr_name);
}

int lrc_attr_get_op_index(int nt_num, int vis_num, char * attr_name) {
  int global_vnum;
  global_vnum=lrc_prod_visnum_vpr_num(lrc_nont_first_prod(nt_num), vis_num);
  return lookup_in_packet_attrs((*visit_table)[global_vnum].op_attrs,attr_name);
}


/** HACK ALERT **/
/** We store the data about a visit to a nonterminal with its first prod !!!! */
void lrc_define_packet_attr
(
  int nont_number, int vis_num, char * attr_name, int packet_index, int is_input
)
{
  int glob_vnum;
  int prod;
 
  prod=lrc_nont_first_prod(nont_number);

  glob_vnum=lrc_prod_visnum_vpr_num(prod, vis_num);

  AssertI(is_vpr_index(glob_vnum),
          "lrc_define_packet_attr:range error:glob_vnum",glob_vnum)

  if (is_input) {
     (*visit_table)[glob_vnum].ip_attrs
     = enter_attr_description((*visit_table)[glob_vnum].ip_attrs,attr_name,packet_index);
  }
  else {
     (*visit_table)[glob_vnum].op_attrs
     = enter_attr_description((*visit_table)[glob_vnum].op_attrs,attr_name,packet_index);
  }
}

int lrc_vpr_index(lrc_ptr f) {
  int ix;
  int found;

  Assert(num_allocs==1,"visit table not allocated")
  ix=0; found=0;
  while ((ix<num_vpr_routines) & !found) {
    found=((*visit_table)[ix].routine==f);
    if (!found) ix++;
  }
  if (found) return ix;
  AssertI(found,"lrc_vpr_index; function not found", (int)f)
  exit(1);
}

/* Access to information on visit routines for PRODUCTIONS */


int lrc_prod_visnum_vpr_num(int prod, int vis_num) {
  /* printf("lrc_prod_visnum_vpr_num\n");
    lrc_print_prod_info(prod);
    lrc_print_nont_info(lrc_prod_lhs_nont(prod));
  */
  AssertI(vis_num<=lrc_nont_num_visits(lrc_prod_lhs_nont(prod)),
          "lrc_prod_visnum_vpr_num:range error:vis_num",vis_num)
  return vis_num-1+lrc_prod_first_visit_index(prod);
}

lrc_ptr lrc_vpr_routine(int vpr_number) {
  Assert(num_allocs==1,"visit table not allocated")
  AssertI(is_vpr_index(vpr_number),"lrc_vpr_routine:range error:vis_number",vpr_number)
  return (*visit_table)[vpr_number].routine;
}
int lrc_vpr_prod_number(int vpr_number) {
  Assert(num_allocs==1,"visit table not allocated")
  AssertI(is_vpr_index(vpr_number),"lrc_vpr_prod_number:range error:vis_number",vpr_number)
  return (*visit_table)[vpr_number].production_number;
}
int lrc_vpr_visit_number(int vpr_number) {
  Assert(num_allocs==1,"visit table not allocated")
  AssertI(is_vpr_index(vpr_number),"lrc_vpr_visit_number:range error:vis_number",vpr_number)
  return (*visit_table)[vpr_number].visit_number;
}
int lrc_vpr_num_ip_attrs(int vpr_number) {
  Assert(num_allocs==1,"visit table not allocated")
  AssertI(is_vpr_index(vpr_number),"lrc_vpr_num_ip_attrs:range error:vis_number",vpr_number)
  return (*visit_table)[vpr_number].num_ip_attrs;
}
int lrc_vpr_num_op_attrs(int vpr_number) {
  Assert(num_allocs==1,"visit table not allocated")
  AssertI(is_vpr_index(vpr_number),"lrc_vpr_num_op_attrs:range error:vis_number",vpr_number)
  return (*visit_table)[vpr_number].num_op_attrs;
}

/* Return sizes of input and output packets */
int lrc_prod_vnum_ip_size(int tag, int visit_number) {
  int vpr_index;

  AssertI(visit_number<=lrc_nont_num_visits(lrc_prod_lhs_nont(tag)),
          "lrc_prod_vnum_ip_size:range error:visit_number",visit_number)
  vpr_index=lrc_prod_visnum_vpr_num(tag,visit_number);
  return lrc_vpr_num_ip_attrs(vpr_index)+1;
}

int lrc_prod_vnum_op_size(int tag, int visit_number) {
  int vpr_index;

  AssertI(visit_number<=lrc_nont_num_visits(lrc_prod_lhs_nont(tag)),
          "lrc_prod_vnum_ip_size:range error:visit_number",visit_number)
  vpr_index=lrc_prod_visnum_vpr_num(tag,visit_number);
  return lrc_vpr_num_op_attrs(vpr_index);
}

/* Access to information on visit routines for NONTERMINALS */

lrc_ptr lrc_vnt_routine(int vnt_number) {
  Assert(num_allocs==1,"visit table not allocated")
  AssertI(is_vnt_index(vnt_number),"lrc_vnt_routine:range error:vnt_number",vnt_number)
  return (*nont_visit_table)[vnt_number].routine;
}

/* return vnt_number of visit-routine for visit vis_num to nonterminal nont */
int lrc_nont_visnum_vnt_num(int nont, int vis_num) {
  AssertI(vis_num<=lrc_nont_num_visits(nont),
          "lrc_nont_visnum_vnt_num:range error:vis_num",vis_num)
  return vis_num-1+lrc_nont_first_visit_index(nont);
}

int lrc_vnt_nont_number(int vnt_number) {
  Assert(num_allocs==1,"visit table not allocated")
  AssertI(is_vnt_index(vnt_number),"lrc_vnt_nont_number:range error:vnt_number",vnt_number)
  return (*nont_visit_table)[vnt_number].nont_number;
}

int lrc_vnt_visit_number(int vnt_number) {
  Assert(num_allocs==1,"visit table not allocated")
  AssertI(is_vnt_index(vnt_number),"lrc_vnt_visit_number:range error:vnt_number",vnt_number)
  return (*nont_visit_table)[vnt_number].visit_number;
}

int lrc_vnt_num_ip_attrs(int vnt_number) {
  Assert(num_allocs==1,"visit table not allocated")
  AssertI(is_vnt_index(vnt_number),"lrc_vnt_num_ip_attrs:range error:vnt_number",vnt_number)
  return (*nont_visit_table)[vnt_number].num_ip_attrs;
}

int lrc_vnt_num_op_attrs(int vnt_number) {
  Assert(num_allocs==1,"visit table not allocated")
  AssertI(is_vnt_index(vnt_number),"lrc_vnt_num_op_attrs:range error:vnt_number",vnt_number)
  return (*nont_visit_table)[vnt_number].num_op_attrs;
}

/* Creating input and output packets */

lrc_term * lrc_create_ip(lrc_term t, int vnum) {
  lrc_term * ip;
  ip=(lrc_term *)lrc_malloc(sizeof(lrc_term)*lrc_prod_vnum_ip_size(LRC_TAG(t),vnum));
  ip[0]=t;
  return ip;
}

lrc_term * lrc_create_op(lrc_term t, int vnum) {
  lrc_term * op;
  op=(lrc_term *)lrc_malloc(sizeof(lrc_term)*lrc_prod_vnum_op_size(LRC_TAG(t),vnum));
  return op;
}

static int do_print_packets=0;

void set_packet_printing()   { do_print_packets=1;}
void unset_packet_printing() { do_print_packets=0;}


/* JAS: Huge hack - global variable to pass the nt number and visnum to 
 *                  print_packet
 */

int jas_vis_num ;
int jas_nt_num  ;
int jas_ipOrop  ; /* 1 is ip , 0 is op */

/* */

static void print_packet(int byte_size, lrc_term *packet, int first_index) {
  int num_elements;
  int ix;
  printf("[");
  num_elements=byte_size/sizeof(lrc_term);
  for(ix=first_index;ix<num_elements;ix++) {
     /* lrc_term_print_address(packet[ix]); printf(";"); */
     /* JAS */

     fprintf(stdout,"JAS: ip: %d ntnum: %d   visnum %d    num_elements: %d  ix: %d ntident: %s JAS\n",jas_ipOrop,jas_nt_num,jas_vis_num,num_elements,ix,lrc_nont_ident(jas_nt_num));fflush(stdout);

     if (jas_ipOrop) 
          printf("%s :",lrc_attr_get_ip_name(jas_nt_num,jas_vis_num,ix));
     else 
          printf("%s :",lrc_attr_get_op_name(jas_nt_num,jas_vis_num,ix));
 
     lrc_term_print_long(packet[ix]); printf(";");
  }
  printf("]");fflush(stdout);
}

/* this function performs the vis_num-th visit to the term in ip[0] */
int lrc_visit_with_packets(int vis_num, lrc_term *ip, lrc_term *op) {
  lrc_term term;
  int vpr_index;
  int ip_size, op_size;
  int res;

  term=ip[0];
  /* printf("Visiting %s\\%d\n", lrc_constructor_ident(LRC_TAG(term)), vis_num); */
  vpr_index=lrc_prod_visnum_vpr_num(LRC_TAG(term),vis_num);
  ip_size=(lrc_vpr_num_ip_attrs(vpr_index)+1)*sizeof(lrc_term);
  op_size=lrc_vpr_num_op_attrs(vpr_index)*sizeof(lrc_term);
  if (do_print_packets) {
     printf("START %s\\%d @%p ", lrc_constructor_ident(LRC_TAG(term)), vis_num, term);

     /* JAS */
     jas_ipOrop = 1;
     jas_nt_num = lrc_prod_lhs_nont(LRC_TAG(term));
     jas_vis_num = vis_num ;

     print_packet(ip_size,ip,1);
     printf("\n"); fflush(stdout);
  }
  res=
    lrc_memo((lrc_visfun) lrc_vpr_routine(vpr_index),
             ip_size, (lrc_packet)ip,
             op_size, (lrc_packet)op
    );
  if (do_print_packets) {
     if (res) printf("HIT "); else printf("END "); 
     /* JAS */
     printf("%s\\%d @%p ", lrc_constructor_ident(LRC_TAG(term)), vis_num, term); 
     jas_ipOrop = 0;
     jas_nt_num = lrc_prod_lhs_nont(LRC_TAG(term));
     jas_vis_num = vis_num ;

     if (!res) print_packet(op_size,op,0);
     printf("\n");
  }
  return res;
}

/* Return the sizes of input and output packet for the visit_number'th
 * visit to term t
 * The sizes are in number of lrc_term's in the packets.
 * The sizes are returned in the parameters ip_size and op_size.
 *
 * The routine returns 1 if all is well, otherwise it returns 0.
 */

int lrc_visit_size(lrc_term t, int visit_number, int *ip_size, int *op_size) {
  int tag;
  int vpr_index;

  tag=LRC_TAG(t);
  if (lrc_prod_num_visits(tag)<visit_number) return 0;
  vpr_index=lrc_prod_visnum_vpr_num(tag,visit_number);
  *ip_size=lrc_vpr_num_ip_attrs(vpr_index)+1;
  *op_size=lrc_vpr_num_op_attrs(vpr_index);
  return 1;
}

/* A general visit routine ...
 *
 *   Parameters: The term t that must be visited and a 0-terminated
 *    list of attribute-values.
 *
 *   This routine performs the first visit to tree t. It calls the
 *   visit routine associated with the nonterminal that derives t.
 *   It returns a pointer to the output packet.
 *
 *   If the number of attribute-values in the call differs
 *   from the number of inherited attributes in the visit then the
 *   visit is not performed and a null pointer is returned.
 */

lrc_term * lrc_visit_term_with_attrs_1(lrc_term t, ...) {
  va_list ap;
  int ip_size, op_size;

  if (lrc_visit_size(t,1,&ip_size,&op_size)) {
    lrc_term *ip;
    lrc_term *op;
    lrc_term attr;
    int i;

    ip= (lrc_term *) lrc_malloc(ip_size * sizeof(lrc_term));
    op= (lrc_term *) lrc_malloc(op_size * sizeof(lrc_term));
    ip[0]=t;

    va_start(ap,t);
    for(i=1;i<ip_size;i++) { /* copy attribute values into input packet */
      attr=va_arg(ap,lrc_term);
      if (attr==0) {
        printf("Missing attributes in lrc_visit_term_with_attrs_1\n");
        return 0;
      }
      ip[i]=attr;
    }
    attr=va_arg(ap,lrc_term);
    if (attr!=0) {
      printf("Too many attributes in call to lrc_visit_term_with_attrs_1\n");
      return 0;
    }
    va_end(ap);
    lrc_visit_with_packets(1,ip,op);
    return op;
  }

  return 0;
}

lrc_term * lrc_visit_term_1(lrc_term t) {
  return lrc_visit_term_with_attrs_1(t,0);
}

lrc_term * lrc_visit_term_with_attr_1(lrc_term t, lrc_term attr) {
  return lrc_visit_term_with_attrs_1(t,attr,0);
}

void lrc_block_copy(lrc_term * dst, int d_offset, lrc_term *src, int s_offset, int num_terms) {
  memcpy(dst+d_offset,src+s_offset,num_terms*sizeof(lrc_term));
}

/*
** A routine for visiting the input tree in an EDSE environment.
** Added by Dick Heijink feb 1996
*/

lrc_term * lrc_visit_edse_root(lrc_term t,lrc_term bindings)
{
  int ip_size, op_size;

  /*** JAS
     printf("lrc_visit_edse_root:TREE="); lrc_term_print_long(t); printf("\n");
     lrc_print_prod_info(LRC_TAG(t));
     lrc_print_nont_info(lrc_prod_lhs_nont(LRC_TAG(t)));
  ***/
  
/********** JAS 
  if (bindings==0) {
     printf("lrc_visit_edse_root: bindings is 0\n");
  }
  else {
     printf("Bindings in are\n");
     lrc_term_print_long(bindings); printf("\n");
  }
**********/
  if (lrc_visit_size(t,1,&ip_size,&op_size)) {
    lrc_term *ip;
    lrc_term *op;
    int bind_in_index;

    /** Allocate input and output packet. */
    ip= (lrc_term *) lrc_malloc(ip_size * sizeof(lrc_term));
    op= (lrc_term *) lrc_malloc(op_size * sizeof(lrc_term));

    /* Initialize input packet. */
    /***** JAS
        printf("Input packet size is %d\n", ip_size);
        printf("Output packet size is %d\n", op_size);
    ******/

    ip[0]=t; /** First position is input tree. */
    if (bindings!=0) {
      bind_in_index = lrc_attr_get_ip_index( lrc_prod_lhs_nont(LRC_TAG(t)), 1, "_lrc_binds_in" );
      if (bind_in_index!=-1) {
          /** printf("Setting ip[%d] to bindings\n", bind_in_index); **/
          ip[bind_in_index]=bindings;
      }
    }

    /* Determine visitfunction and call it. */

    lrc_memo_start_AE();
    /** printf("lrc_visit_edse_root:visiting the tree...\n"); **/
    lrc_visit_with_packets(1,ip,op);
    lrc_memo_end_AE();
    /** printf("lrc_visit_edse_root:done with visit\n"); **/
    return op;
  }
  /* printf("lrc_visit_edse_root:NOT visiting the tree\n"); */
  return 0;
}




